import { AbstractService, LegacyApiServiceInterface, InternalEventBusInterface, IntegrityApiInterface, ItemsServerInterface, ApiServiceEvent, KeyValueStoreInterface, ApiServiceEventData } from '@standardnotes/services';
import { DownloadFileParams, FileOwnershipType, FilesApiInterface } from '@standardnotes/files';
import { ServerSyncPushContextualPayload, SNFeatureRepo } from '@standardnotes/models';
import { User, HttpStatusCode, KeyParamsResponse, SignInResponse, SignOutResponse, ChangeCredentialsResponse, RawSyncResponse, SessionRenewalResponse, SessionListResponse, ListSettingsResponse, UpdateSettingResponse, GetSettingResponse, DeleteSettingResponse, ListedRegistrationResponse, StartUploadSessionResponse, IntegrityPayload, CheckIntegrityResponse, GetSingleItemResponse, HttpResponse, HttpResponseMeta, ErrorTag, HttpRequest, ClientDisplayableError, HttpErrorResponse, ValetTokenOperation, MetaEndpointResponse } from '@standardnotes/responses';
import { LegacySession, MapperInterface, Session } from '@standardnotes/domain-core';
import { HttpServiceInterface } from '@standardnotes/api';
import { SNRootKeyParams } from '@standardnotes/encryption';
import { PureCryptoInterface } from '@standardnotes/sncrypto-common';
import { DiskStorageService } from '../Storage/DiskStorageService';
import { UuidString } from '../../Types/UuidString';
import { SettingsServerInterface } from '../Settings/SettingsServerInterface';
import { AnyFeatureDescription } from '@standardnotes/features';
type InvalidSessionObserver = (revoked: boolean) => void;
export declare class LegacyApiService extends AbstractService<ApiServiceEvent, ApiServiceEventData> implements LegacyApiServiceInterface, FilesApiInterface, IntegrityApiInterface, ItemsServerInterface, SettingsServerInterface {
    private httpService;
    private storageService;
    private host;
    private inMemoryStore;
    private crypto;
    private sessionStorageMapper;
    private legacySessionStorageMapper;
    protected internalEventBus: InternalEventBusInterface;
    private session;
    user?: User;
    private authenticating;
    private changing;
    private refreshingSession;
    private invalidSessionObserver?;
    private filesHost?;
    constructor(httpService: HttpServiceInterface, storageService: DiskStorageService, host: string, inMemoryStore: KeyValueStoreInterface<string>, crypto: PureCryptoInterface, sessionStorageMapper: MapperInterface<Session, Record<string, unknown>>, legacySessionStorageMapper: MapperInterface<LegacySession, Record<string, unknown>>, internalEventBus: InternalEventBusInterface);
    deinit(): void;
    setUser(user?: User): void;
    /**
     * When a we receive a 401 error from the server, we'll notify the observer.
     * Note that this applies only to sessions that are totally invalid. Sessions that
     * are expired but can be renewed are still considered to be valid. In those cases,
     * the server response is 498.
     * If the session has been revoked, then the observer will have its first
     * argument set to true.
     */
    setInvalidSessionObserver(observer: InvalidSessionObserver): void;
    loadHost(): string;
    setHost(host: string): Promise<void>;
    getHost(): string;
    getFilesHost(): string;
    setSession(session: Session | LegacySession, persist?: boolean): void;
    getSession(): Session | LegacySession | null;
    get apiVersion(): string;
    private params;
    createErrorResponse(message: string, status?: HttpStatusCode, tag?: ErrorTag): HttpErrorResponse;
    private errorResponseWithFallbackMessage;
    processMetaObject(meta: HttpResponseMeta): void;
    private processSuccessResponseForMetaBody;
    private request;
    /**
     * @param mfaCode     The mfa challenge response value.
     */
    getAccountKeyParams(dto: {
        email: string;
        mfaCode?: string;
        authenticatorResponse?: Record<string, unknown>;
    }): Promise<HttpResponse<KeyParamsResponse>>;
    signIn(dto: {
        email: string;
        serverPassword: string;
        ephemeral: boolean;
        hvmToken?: string;
    }): Promise<HttpResponse<SignInResponse>>;
    signOut(): Promise<HttpResponse<SignOutResponse>>;
    changeCredentials(parameters: {
        userUuid: UuidString;
        currentServerPassword: string;
        newServerPassword: string;
        newKeyParams: SNRootKeyParams;
        newEmail?: string;
    }): Promise<HttpResponse<ChangeCredentialsResponse>>;
    sync(payloads: ServerSyncPushContextualPayload[], lastSyncToken: string | undefined, paginationToken: string | undefined, limit: number, sharedVaultUuids?: string[]): Promise<HttpResponse<RawSyncResponse>>;
    getSyncHttpRequest(payloads: ServerSyncPushContextualPayload[], lastSyncToken: string | undefined, paginationToken: string | undefined, limit: number, sharedVaultUuids?: string[] | undefined): HttpRequest;
    /**
     * @deprecated
     *
     * This function should be replaced with @standardnotes/api's `HttpService::refreshSession` function.
     */
    deprecatedRefreshSessionOnlyUsedInE2eTests(): Promise<HttpResponse<SessionRenewalResponse>>;
    getSessionsList(): Promise<HttpResponse<SessionListResponse>>;
    deleteSession(sessionId: UuidString): Promise<HttpResponse<SessionListResponse>>;
    private tokenRefreshableRequest;
    listSettings(userUuid: UuidString): Promise<HttpResponse<ListSettingsResponse>>;
    updateSetting(userUuid: UuidString, settingName: string, settingValue: string | null, sensitive: boolean): Promise<HttpResponse<UpdateSettingResponse>>;
    getSetting(userUuid: UuidString, settingName: string): Promise<HttpResponse<GetSettingResponse>>;
    getSubscriptionSetting(userUuid: UuidString, settingName: string): Promise<HttpResponse<GetSettingResponse>>;
    updateSubscriptionSetting(userUuid: UuidString, settingName: string, settingValue: string | null, sensitive: boolean): Promise<HttpResponse<UpdateSettingResponse>>;
    deleteSetting(userUuid: UuidString, settingName: string): Promise<HttpResponse<DeleteSettingResponse>>;
    downloadFeatureUrl(url: string): Promise<HttpResponse>;
    getNewSubscriptionToken(): Promise<string | undefined>;
    downloadOfflineFeaturesFromRepo(dto: {
        repo: SNFeatureRepo;
    }): Promise<{
        features: AnyFeatureDescription[];
        roles: string[];
    } | ClientDisplayableError>;
    registerForListedAccount(): Promise<HttpResponse<ListedRegistrationResponse>>;
    createUserFileValetToken(remoteIdentifier: string, operation: ValetTokenOperation, unencryptedFileSize?: number): Promise<string | ClientDisplayableError>;
    startUploadSession(valetToken: string, ownershipType: FileOwnershipType): Promise<HttpResponse<StartUploadSessionResponse>>;
    deleteFile(valetToken: string, ownershipType: FileOwnershipType): Promise<HttpResponse<StartUploadSessionResponse>>;
    uploadFileBytes(valetToken: string, ownershipType: FileOwnershipType, chunkId: number, encryptedBytes: Uint8Array): Promise<boolean>;
    closeUploadSession(valetToken: string, ownershipType: FileOwnershipType): Promise<boolean>;
    moveFile(valetToken: string): Promise<boolean>;
    getFilesDownloadUrl(ownershipType: FileOwnershipType): string;
    downloadFile({ file, chunkIndex, valetToken, ownershipType, contentRangeStart, onBytesReceived, }: DownloadFileParams): Promise<ClientDisplayableError | undefined>;
    checkIntegrity(integrityPayloads: IntegrityPayload[]): Promise<HttpResponse<CheckIntegrityResponse>>;
    getSingleItem(itemUuid: string): Promise<HttpResponse<GetSingleItemResponse>>;
    private preprocessingError;
    private preprocessAuthenticatedErrorResponse;
    private getSessionAccessToken;
    getCaptchaUrl(): Promise<HttpResponse<MetaEndpointResponse>>;
}
export {};
