import { AbstractService, InternalEventBusInterface, MutatorClientInterface, AlertService } from '@standardnotes/services';
import { ItemManager } from '../Items';
import { PayloadManager } from '../Payloads/PayloadManager';
import { ActionsExtensionMutator, ComponentInterface, ComponentMutator, DecryptedItemInterface, DecryptedItemMutator, DecryptedPayloadInterface, EncryptedItemInterface, FeatureRepoMutator, FileItem, ItemContent, ItemsKeyInterface, ItemsKeyMutatorInterface, MutationType, NoteMutator, PayloadEmitSource, PredicateInterface, SmartView, SNActionsExtension, SNFeatureRepo, SNNote, SNTag, TagMutator, TransactionalMutation, VaultListingInterface } from '@standardnotes/models';
export declare class MutatorService extends AbstractService implements MutatorClientInterface {
    private itemManager;
    private payloadManager;
    private alerts;
    protected internalEventBus: InternalEventBusInterface;
    constructor(itemManager: ItemManager, payloadManager: PayloadManager, alerts: AlertService, internalEventBus: InternalEventBusInterface);
    deinit(): void;
    /**
     * Consumers wanting to modify an item should run it through this block,
     * so that data is properly mapped through our function, and latest state
     * is properly reconciled.
     */
    changeItem<M extends DecryptedItemMutator = DecryptedItemMutator, I extends DecryptedItemInterface = DecryptedItemInterface>(itemToLookupUuidFor: I, mutate?: (mutator: M) => void, mutationType?: MutationType, emitSource?: PayloadEmitSource, payloadSourceKey?: string): Promise<I>;
    /**
     * @param mutate If not supplied, the intention would simply be to mark the item as dirty.
     */
    changeItems<M extends DecryptedItemMutator = DecryptedItemMutator, I extends DecryptedItemInterface = DecryptedItemInterface>(itemsToLookupUuidsFor: I[], mutate?: (mutator: M) => void, mutationType?: MutationType, emitSource?: PayloadEmitSource, payloadSourceKey?: string): Promise<I[]>;
    /**
     * Run unique mutations per each item in the array, then only propagate all changes
     * once all mutations have been run. This differs from `changeItems` in that changeItems
     * runs the same mutation on all items.
     */
    runTransactionalMutations(transactions: TransactionalMutation[], emitSource?: PayloadEmitSource, payloadSourceKey?: string): Promise<(DecryptedItemInterface | undefined)[]>;
    runTransactionalMutation(transaction: TransactionalMutation, emitSource?: PayloadEmitSource, payloadSourceKey?: string): Promise<DecryptedItemInterface | undefined>;
    changeNote(itemToLookupUuidFor: SNNote, mutate: (mutator: NoteMutator) => void, mutationType?: MutationType, emitSource?: PayloadEmitSource, payloadSourceKey?: string): Promise<DecryptedPayloadInterface[]>;
    changeTag(itemToLookupUuidFor: SNTag, mutate: (mutator: TagMutator) => void, mutationType?: MutationType, emitSource?: PayloadEmitSource, payloadSourceKey?: string): Promise<SNTag>;
    changeComponent(itemToLookupUuidFor: ComponentInterface, mutate: (mutator: ComponentMutator) => void, mutationType?: MutationType, emitSource?: PayloadEmitSource, payloadSourceKey?: string): Promise<ComponentInterface>;
    changeFeatureRepo(itemToLookupUuidFor: SNFeatureRepo, mutate: (mutator: FeatureRepoMutator) => void, mutationType?: MutationType, emitSource?: PayloadEmitSource, payloadSourceKey?: string): Promise<SNFeatureRepo>;
    changeActionsExtension(itemToLookupUuidFor: SNActionsExtension, mutate: (mutator: ActionsExtensionMutator) => void, mutationType?: MutationType, emitSource?: PayloadEmitSource, payloadSourceKey?: string): Promise<SNActionsExtension>;
    changeItemsKey(itemToLookupUuidFor: ItemsKeyInterface, mutate: (mutator: ItemsKeyMutatorInterface) => void, mutationType?: MutationType, emitSource?: PayloadEmitSource, payloadSourceKey?: string): Promise<ItemsKeyInterface>;
    private applyTransform;
    /**
     * Sets the item as needing sync. The item is then run through the mapping function,
     * and propagated to mapping observers.
     * @param isUserModified - Whether to update the item's "user modified date"
     */
    setItemDirty(itemToLookupUuidFor: DecryptedItemInterface, isUserModified?: boolean): Promise<DecryptedItemInterface<ItemContent>>;
    setItemsDirty(itemsToLookupUuidsFor: DecryptedItemInterface[], isUserModified?: boolean): Promise<DecryptedItemInterface[]>;
    /**
     * Duplicates an item and maps it, thus propagating the item to observers.
     * @param isConflict - Whether to mark the duplicate as a conflict of the original.
     */
    duplicateItem<T extends DecryptedItemInterface>(itemToLookupUuidFor: T, isConflict?: boolean, additionalContent?: Partial<ItemContent>): Promise<T>;
    createItem<T extends DecryptedItemInterface, C extends ItemContent = ItemContent>(contentType: string, content: C, needsSync?: boolean, vault?: VaultListingInterface): Promise<T>;
    insertItem<T extends DecryptedItemInterface>(item: DecryptedItemInterface, setDirty?: boolean): Promise<T>;
    insertItems(items: DecryptedItemInterface[], emitSource?: PayloadEmitSource): Promise<DecryptedItemInterface[]>;
    emitItemFromPayload<T extends DecryptedItemInterface>(payload: DecryptedPayloadInterface, emitSource: PayloadEmitSource): Promise<T>;
    emitItemsFromPayloads(payloads: DecryptedPayloadInterface[], emitSource: PayloadEmitSource): Promise<DecryptedItemInterface[]>;
    setItemToBeDeleted(itemToLookupUuidFor: DecryptedItemInterface | EncryptedItemInterface, source?: PayloadEmitSource): Promise<void>;
    setItemsToBeDeleted(itemsToLookupUuidsFor: (DecryptedItemInterface | EncryptedItemInterface)[]): Promise<void>;
    findOrCreateTagParentChain(titlesHierarchy: string[]): Promise<SNTag>;
    createTag(dto: {
        title: string;
        parentItemToLookupUuidFor?: SNTag;
        createInVault?: VaultListingInterface;
    }): Promise<SNTag>;
    createSmartView<T extends DecryptedItemInterface>(dto: {
        title: string;
        predicate: PredicateInterface<T>;
        iconString?: string;
        vault?: VaultListingInterface;
    }): Promise<SmartView>;
    createSmartViewFromDSL<T extends DecryptedItemInterface>(dsl: string, vault?: VaultListingInterface): Promise<SmartView>;
    createTagOrSmartView<T extends SNTag | SmartView>(title: string, vault?: VaultListingInterface): Promise<T>;
    findOrCreateTagByTitle(dto: {
        title: string;
        parentItemToLookupUuidFor?: SNTag;
        createInVault?: VaultListingInterface;
    }): Promise<SNTag>;
    renameFile(file: FileItem, name: string): Promise<FileItem>;
    mergeItem(item: DecryptedItemInterface, source: PayloadEmitSource): Promise<DecryptedItemInterface>;
    setItemNeedsSync(item: DecryptedItemInterface, updateTimestamps?: boolean): Promise<DecryptedItemInterface | undefined>;
    setItemsNeedsSync(items: DecryptedItemInterface[]): Promise<(DecryptedItemInterface | undefined)[]>;
    deleteItem(item: DecryptedItemInterface | EncryptedItemInterface): Promise<void>;
    deleteItems(items: (DecryptedItemInterface | EncryptedItemInterface)[]): Promise<void>;
    /**
     * Permanently deletes any items currently in the trash. Consumer must manually call sync.
     */
    emptyTrash(): Promise<void>;
    migrateTagsToFolders(): Promise<void>;
    findOrCreateTag(title: string, createInVault?: VaultListingInterface): Promise<SNTag>;
    /**
     * @returns The changed child tag
     */
    setTagParent(parentTag: SNTag, childTag: SNTag): Promise<SNTag>;
    /**
     * @returns The changed child tag
     */
    unsetTagParent(childTag: SNTag): Promise<SNTag>;
    associateFileWithNote(file: FileItem, note: SNNote): Promise<FileItem | undefined>;
    disassociateFileWithNote(file: FileItem, note: SNNote): Promise<FileItem>;
    addTagToNote(note: SNNote, tag: SNTag, addHierarchy: boolean): Promise<SNTag[] | undefined>;
    addTagToFile(file: FileItem, tag: SNTag, addHierarchy: boolean): Promise<SNTag[] | undefined>;
    linkNoteToNote(note: SNNote, otherNote: SNNote): Promise<SNNote>;
    linkFileToFile(file: FileItem, otherFile: FileItem): Promise<FileItem>;
    unlinkItems(itemA: DecryptedItemInterface<ItemContent>, itemB: DecryptedItemInterface<ItemContent>): Promise<DecryptedItemInterface<ItemContent>>;
}
