import { SNRootKey } from '@standardnotes/encryption';
import { AbstractService, StorageServiceInterface, InternalEventHandlerInterface, StoragePersistencePolicies, StorageValuesObject, DeviceInterface, InternalEventBusInterface, InternalEventInterface, StorageValueModes, WrappedStorageValue, ValuesObjectRecord, EncryptionProviderInterface } from '@standardnotes/services';
import { FullyFormedPayloadInterface, DeletedPayloadInterface, FullyFormedTransferPayload } from '@standardnotes/models';
/**
 * The storage service is responsible for persistence of both simple key-values, and payload
 * storage. It does so by relying on deviceInterface to save and retrieve raw values and payloads.
 * For simple key/values, items are grouped together in an in-memory hash, and persisted to disk
 * as a single object (encrypted, when possible). It handles persisting payloads in the local
 * database by encrypting the payloads when possible.
 * The storage service also exposes methods that allow the application to initially
 * decrypt the persisted key/values, and also a method to determine whether a particular
 * key can decrypt wrapped storage.
 */
export declare class DiskStorageService extends AbstractService implements StorageServiceInterface, InternalEventHandlerInterface {
    private device;
    private identifier;
    protected internalEventBus: InternalEventBusInterface;
    private encryptionProvider;
    private storagePersistable;
    private persistencePolicy;
    private needsPersist;
    private currentPersistPromise?;
    private values;
    constructor(device: DeviceInterface, identifier: string, internalEventBus: InternalEventBusInterface);
    provideEncryptionProvider(provider: EncryptionProviderInterface): void;
    deinit(): void;
    handleEvent(event: InternalEventInterface): Promise<void>;
    setPersistencePolicy(persistencePolicy: StoragePersistencePolicies): Promise<void>;
    isEphemeralSession(): boolean;
    initializeFromDisk(): Promise<void>;
    private setInitialValues;
    isStorageWrapped(): boolean;
    canDecryptWithKey(key: SNRootKey): Promise<boolean>;
    private decryptWrappedValue;
    decryptStorage(): Promise<void>;
    /** @todo This function should be debounced. */
    private persistValuesToDisk;
    awaitPersist(): Promise<void>;
    private immediatelyPersistValuesToDisk;
    /**
     * Generates a payload that can be persisted to disk,
     * either as a plain object, or an encrypted item.
     */
    private generatePersistableValues;
    setValue<T>(key: string, value: T, mode?: StorageValueModes): void;
    setValueAndAwaitPersist(key: string, value: unknown, mode?: StorageValueModes): Promise<void>;
    private setValueWithNoPersist;
    getValue<T>(key: string, mode?: StorageValueModes, defaultValue?: T): T;
    getAllKeys(mode?: StorageValueModes): string[];
    removeValue(key: string, mode?: StorageValueModes): Promise<void>;
    /**
     * Default persistence key. Platforms can override as needed.
     */
    private getPersistenceKey;
    private defaultValuesObject;
    static DefaultValuesObject(wrapped?: WrappedStorageValue, unwrapped?: ValuesObjectRecord, nonwrapped?: ValuesObjectRecord): StorageValuesObject;
    private domainKeyForMode;
    /**
     * Clears simple values from storage only. Does not affect payloads.
     */
    clearValues(): Promise<void>;
    getAllRawPayloads(): Promise<FullyFormedTransferPayload[]>;
    getRawPayloads(uuids: string[]): Promise<FullyFormedTransferPayload[]>;
    savePayload(payload: FullyFormedPayloadInterface): Promise<void>;
    savePayloads(payloads: FullyFormedPayloadInterface[]): Promise<void>;
    deletePayloads(payloads: DeletedPayloadInterface[]): Promise<void>;
    deletePayloadsWithUuids(uuids: string[]): Promise<void>;
    deletePayloadWithUuid(uuid: string): Promise<void>;
    clearAllPayloads(): Promise<void>;
    clearAllData(): Promise<void>;
}
